/**
 * Copyright (C) 2007-2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#include <sys/utsname.h>

#include <glib/gstdio.h>
#include <gtk/gtk.h>

#include <libsoup/soup-cookie-jar-sqlite.h>

#include <proxy.h>

#include "main.h"
#include "enums.h"
#include "utils.h"
#include "image-utils.h"
#include "wk-video.h"
#include "wk-events.h"
#include "wk-browser.h"
#include "wk-daemon.h"
#include "wk-window.h"
#include "wk-view.h"
#include "wk-logger.h"

enum
{
    VIEW_PAGE_WEB = 0,
    VIEW_PAGE_IMAGE,
    VIEW_PAGE_VIDEO,
    VIEW_PAGE_BLACK
};

/* signals */
enum
{
    SESSION_START_SIGNAL = 0,
    SESSION_STOP_SIGNAL,
    TIME_ADDED_SIGNAL,
    LAST_SIGNAL
};

static guint fe_signals[LAST_SIGNAL] = { 0 };

gboolean is_granted = TRUE;
gboolean webview_no_switch = FALSE;
guint context_id;

static void _view_build_toolbar   (WkView *view);
static void _view_build_statusbar (WkView *view);

static void _view_on_webkit_load_status (WebKitWebView *webkitwebview,
                                         GParamSpec    *pspec,
                                         WkView        *view);
static void _view_notify_progress       (WebKitWebView *webkitwebview,
                                         GParamSpec    *pspec,
                                         WkView        *view);
static void _view_link_hover            (WebKitWebView *webkitwebview,
                                         const gchar   *title,
                                         const gchar   *link,
                                         WkView        *view);

static WebKitWebView *_view_create_web_view (WebKitWebView  *web_view,
                                             WebKitWebFrame *frame,
                                             gpointer        data);
static gboolean _view_check_mime_policy (WebKitWebView           *web_view,
                                         WebKitWebFrame          *frame,
                                         WebKitNetworkRequest    *request,
                                         gchar                   *mimetype,
                                         WebKitWebPolicyDecision *policy_decision,
                                         WkView                  *view);
static gboolean _view_do_download (WebKitWebView *web_view,
                                   GObject       *download,
                                   WkView        *view);
static void _view_download_notify_status (WebKitDownload *download,
                                          GParamSpec     *pspec,
                                          WkView         *view);

static void _view_on_request_queued (SoupSession *session,
                                     SoupMessage *msg,
                                     WkView      *view);
static void _view_on_auth_requested (SoupSession *session,
                               SoupMessage *msg,
                               SoupAuth    *auth,
                               gboolean     retrying,
                               WkView      *view);
static gboolean _view_uri_access_granted (const gchar *uri);

static gboolean _view_set_webpage (WkView *view, const gchar *uri);

static void _view_go                  (GtkWidget *widget, WkView *view);
static void _view_go_home             (GtkWidget *widget, WkView *view);
static void _view_stop_loading        (GtkWidget *widget, WkView *view);
static void _view_go_back             (GtkWidget *widget, WkView *view);
static void _view_go_forward          (GtkWidget *widget, WkView *view);
static void _view_set_current_view    (GtkWidget *entry, WkView *view);
static gboolean _view_url_entry_modified (GtkWidget   *entry,
                                          GdkEventKey *event,
                                          WkView      *view);
static void _view_set_history_buttons (WkView *view);

static void _view_ask_stop_session (GtkWidget *widget, WkView *view);
static void _view_clear_session    (GtkWidget *widget, WkView *view);

static void _view_empty_webview (WkView *view);

static gboolean _view_decrease_time (WkView *view);

static void _view_on_video_end (WkVideo *video, WkView *view);
static void _view_reinit_video_tab (WkView *view);

struct _WkViewPrivate
{
    /* toolbar */
    GtkWidget   *toolbar;
    GtkWidget   *urlEntry;
    GtkWidget   *timeLabel;
    GtkToolItem *backwardBtn;
    GtkToolItem *forwardBtn;
    GtkToolItem *homeBtn;
    GtkToolItem *stopBtn;
    GtkToolItem *clearBtn;
    GtkToolItem *resetBtn;

    /* statusbar */
    GtkWidget *statusbar;
    GtkWidget *progressbar;

    /* timer */
    guint time;
    guint time_timeout;

    /* global settings */
    WkSettings *settings;

    /* Soup */
    SoupSession *session;

    /* Misc */
    WkViewState state;
    gboolean    entry_edited;
    GtkWidget  *black_img;

    /* webview */
    gchar *uri;
};

static void wk_view_class_init (WkViewClass *klass);
static void wk_view_init (WkView *view);
static void wk_view_finalize (GObject *object);

G_DEFINE_TYPE (WkView, wk_view, GTK_TYPE_NOTEBOOK);

static void
wk_view_class_init (WkViewClass *klass)
{
    g_type_class_add_private (klass, sizeof (WkViewPrivate));

    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = wk_view_finalize;

    fe_signals[SESSION_START_SIGNAL] = g_signal_new ("session-start",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (WkViewClass, session_start),
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);

    fe_signals[SESSION_STOP_SIGNAL] = g_signal_new ("session-stop",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (WkViewClass, session_end),
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);

    fe_signals[TIME_ADDED_SIGNAL] = g_signal_new ("time-added",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (WkViewClass, time_added),
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);
}

static void
wk_view_init (WkView *view)
{
    pxProxyFactory *pf;
    char **proxies, **p;
    GtkWidget   *sc, *vbox;
    gchar *user_agent = NULL;
    gchar *sys_name = NULL;
    struct utsname name;
    WebKitWebSettings *web_settings;
    SoupURI     *soup_proxy_uri;

    DEBUG ("creating new view");

    vbox = gtk_vbox_new (FALSE, 0);

    view->priv = G_TYPE_INSTANCE_GET_PRIVATE (view,
                                              WK_TYPE_VIEW,
                                              WkViewPrivate);

    view->priv->settings = GET_SETTINGS ();

    view->priv->time = 0;
    view->priv->time_timeout = 0;

    view->priv->state = WK_VIEW_STATE_NONE;
    view->priv->entry_edited = false;

    view->priv->uri = NULL;

    _view_build_toolbar (view);
    _view_build_statusbar (view);

    gtk_box_pack_start (GTK_BOX (vbox), view->priv->toolbar, FALSE, FALSE, 0);

    sc = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox), sc, TRUE, TRUE, 0);
    gtk_widget_show (sc);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sc),
                                    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    view->webview = WEBKIT_WEB_VIEW (webkit_web_view_new ());
    g_signal_connect (G_OBJECT (view->webview), "notify::load-status",
                      G_CALLBACK (_view_on_webkit_load_status), view);
    g_signal_connect (G_OBJECT (view->webview), "notify::progress",
                      G_CALLBACK (_view_notify_progress), view);
    g_signal_connect (G_OBJECT (view->webview), "hovering-over-link",
                      G_CALLBACK (_view_link_hover), view);
    g_signal_connect (view->webview, "mime-type-policy-decision-requested",
                      G_CALLBACK (_view_check_mime_policy), view);
    g_signal_connect (view->webview, "create-web-view",
                      G_CALLBACK (_view_create_web_view), view);
    g_signal_connect (view->webview, "download-requested",
                      G_CALLBACK (_view_do_download), view);
    gtk_container_add (GTK_CONTAINER (sc), GTK_WIDGET (view->webview));

    gtk_box_pack_start (GTK_BOX (vbox), view->priv->statusbar, FALSE, FALSE, 0);

    view->image = gtk_image_new ();
    view->video = wk_video_new (view->priv->settings->slideshow_mute);
    g_signal_connect (G_OBJECT (view->video), "end-reached",
                      G_CALLBACK (_view_on_video_end), view);

    view->priv->session = webkit_get_default_session ();

    /* remove the webkit auth dialog */
    soup_session_remove_feature_by_type (view->priv->session,
                                    g_type_from_name ("WebKitSoupAuthDialog"));
    g_signal_connect (G_OBJECT (view->priv->session), "request-queued",
                      G_CALLBACK (_view_on_request_queued), view);

    /* automatically detect language settings */
    g_object_set (view->priv->session, "accept-language-auto", TRUE, NULL);

    /* if not done, define the user agent, borrowed from midori */
    if (!view->priv->settings->user_agent)
    {
        if (uname (&name) != -1)
            sys_name = g_strdup(name.sysname);
        else
            sys_name = g_strdup ("Unix");
        user_agent = g_strdup_printf ("Mozilla/5.0 (X11; U; %s; %s) "
                                      "AppleWebKit/%d.%d (KHTML, like Gecko) "
                                      "LinutopKiosk/%s Safari/%d.%d",
                       sys_name,
                       pango_language_to_string (gtk_get_default_language ()),
                       WEBKIT_USER_AGENT_MAJOR_VERSION, WEBKIT_USER_AGENT_MINOR_VERSION,
                       PACKAGE_VERSION,
                       WEBKIT_USER_AGENT_MAJOR_VERSION, WEBKIT_USER_AGENT_MINOR_VERSION);
    }
    else
    {
        user_agent = g_strdup (view->priv->settings->user_agent);
    }
    g_object_set (view->priv->session, "user-agent", user_agent, NULL);

    web_settings = webkit_web_view_get_settings (view->webview);
    g_object_set (web_settings,
                  "WebKitWebSettings::enable-scripts", TRUE,
                  "WebKitWebSettings::user-agent", user_agent,
                  "default-font-size", 11, /* FIXME: setup a setting for this */
                  NULL);

    g_free (sys_name);
    g_free (user_agent);

    /* set the proxy URI if any */
    g_setenv ("PX_CONFIG_ORDER", "config_envvar", FALSE);
    pf = px_proxy_factory_new ();
    proxies = px_proxy_factory_get_proxies (pf, "http://www.linutop.com");

    if (proxies[0] && (strcmp (proxies[0], "direct://") != 0))
    {
        DEBUG ("proxy URI: %s", proxies[0]);
        soup_proxy_uri = soup_uri_new (proxies[0]);
        if (g_getenv ("http_proxy_user"))
            soup_uri_set_user (soup_proxy_uri, g_getenv ("http_proxy_user"));
        if (g_getenv ("http_proxy_password"))
            soup_uri_set_password (soup_proxy_uri, g_getenv ("http_proxy_password"));
        g_object_set (view->priv->session, "proxy-uri", soup_proxy_uri, NULL);
        soup_uri_free (soup_proxy_uri);
    }
    for (p=proxies;*p;p++)
        g_free (*p);
    g_free (proxies);

    /* notebook properties */
    g_object_set (G_OBJECT (view), "show-border", FALSE,
                                   "show-tabs", FALSE,
                                   NULL);

    view->priv->black_img = gtk_image_new ();

    gtk_notebook_append_page (GTK_NOTEBOOK (view), vbox, NULL);
    gtk_notebook_append_page (GTK_NOTEBOOK (view), view->image, NULL);
    gtk_notebook_append_page (GTK_NOTEBOOK (view), view->video, NULL);
    gtk_notebook_append_page (GTK_NOTEBOOK (view), view->priv->black_img, NULL);

    gtk_widget_show_all (GTK_WIDGET (view));

    wk_view_show_toolbar (view, FALSE);
    wk_view_show_statusbar (view, FALSE);
}

static void
wk_view_finalize (GObject *object)
{
    WkView *view = WK_VIEW (object);
    GSource *source;

    if (view->priv->time_timeout)
    {
        if ((source = g_main_context_find_source_by_id (
                          NULL, view->priv->time_timeout)))
            g_source_destroy (source);
    }

    (*G_OBJECT_CLASS (wk_view_parent_class)->finalize) (object);
}



/*
 * Internal callbacks
 */
static void
_view_on_popup_status_changed (WebKitWebView *webview,
                               GParamSpec    *spec,
                               gpointer       data)
{
    const gchar *uri;
    GtkWidget *window;
    WkBrowser *browser;

    DEBUG ("load status: %d", webkit_web_view_get_load_status (webview));

    if (webkit_web_view_get_load_status (webview) == WEBKIT_LOAD_PROVISIONAL)
        return;

    browser = WK_BROWSER (wk_browser_get (NULL));
    window = GTK_WIDGET (data);
    uri = webkit_web_view_get_uri (webview);

    DEBUG ("popup status uri: %s", uri);

    if (window != wk_browser_get_main_window (browser))
    {
        if (! uri || ! _view_uri_access_granted (uri))
        {
            wk_browser_unregister_window (browser, WK_WINDOW (window));
            gtk_widget_destroy (window);
            return;
        }
    }

    gtk_widget_show (window);

    /* this callback is used only once */
    g_signal_handlers_disconnect_by_func (webview, _view_on_popup_status_changed, data);
}

static void
_view_on_webkit_load_status (WebKitWebView *webkitwebview,
                             GParamSpec    *spec,
                             WkView        *view)
{
    const gchar *uri;
    WebKitWebFrame *frame;

    switch (webkit_web_view_get_load_status (view->webview))
    {
    case WEBKIT_LOAD_COMMITTED:
        gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), TRUE);
        frame = webkit_web_view_get_main_frame (view->webview);
        uri = webkit_web_frame_get_uri (frame);
        if (uri)
            gtk_entry_set_text (GTK_ENTRY (view->priv->urlEntry), uri);
        break;

    case WEBKIT_LOAD_PROVISIONAL:
    case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
        gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), TRUE);
        break;

    case WEBKIT_LOAD_FINISHED:
        if (webview_no_switch)
        {
            webview_no_switch = FALSE;
            return;
        }
        gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), FALSE);
        uri = webkit_web_view_get_uri (webkitwebview);
        if (uri && ! view->priv->entry_edited)
        {
            gtk_entry_set_text (GTK_ENTRY (view->priv->urlEntry), uri);
            view->priv->entry_edited = false;
        }
        gtk_notebook_set_current_page (GTK_NOTEBOOK (view), VIEW_PAGE_WEB);
        gdk_flush ();
        break;

    default:
        gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), FALSE);
        break;
    }
    _view_set_history_buttons (view);
}

static void
_view_notify_progress (WebKitWebView *webkitwebview,
                       GParamSpec    *pspec,
                       WkView        *view)
{
    gdouble progress;

    progress = webkit_web_view_get_progress (view->webview);

    if (progress == 1.0)
        gtk_widget_hide (view->priv->progressbar);
    else
    {
        gtk_widget_show (view->priv->progressbar);
        gtk_progress_bar_set_fraction (
            GTK_PROGRESS_BAR (view->priv->progressbar), progress);
    }
}

static void
_view_link_hover (WebKitWebView *webkitwebview,
                  const gchar   *title,
                  const gchar   *link,
                  WkView        *view)
{
    gtk_statusbar_pop (GTK_STATUSBAR (view->priv->statusbar), context_id);
    if (link)
        gtk_statusbar_push (GTK_STATUSBAR (view->priv->statusbar),
                            context_id, link);
}

static void
_view_empty_webview (WkView *view)
{
    const gchar *html = "<html></html>";

    webview_no_switch = TRUE;

    webkit_web_view_load_string (view->webview, html, NULL, NULL, "");
}

static void
_view_set_history_buttons (WkView *view)
{
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->backwardBtn),
        webkit_web_view_can_go_back (view->webview));
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->forwardBtn),
        webkit_web_view_can_go_back_or_forward (view->webview, 1));
}

static void
_view_go (GtkWidget *widget, WkView *view)
{
    gtk_widget_activate (view->priv->urlEntry);
}

static void
_view_stop_loading (GtkWidget *widget, WkView *view)
{
    webkit_web_view_stop_loading (view->webview);
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), FALSE);
}

static void
_view_go_home (GtkWidget *widget, WkView *view)
{
    _view_set_webpage (view, view->priv->settings->default_url);
}

static void
_view_go_back (GtkWidget *widget, WkView *view)
{
    webkit_web_view_go_back (view->webview);
    _view_set_history_buttons (view);
}

static void
_view_go_forward (GtkWidget *widget, WkView *view)
{
    webkit_web_view_go_forward (view->webview);
    _view_set_history_buttons (view);
}

static void
_view_ask_stop_session (GtkWidget *widget, WkView *view)
{
    GtkWidget *dialog;
    gint response;

    WkDaemon *d = wk_daemon_get ();

    dialog = gtk_message_dialog_new (NULL,
                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_QUESTION,
                            GTK_BUTTONS_YES_NO,
                            _("Would you really like to end this session?"));
    response = gtk_dialog_run (GTK_DIALOG (dialog));

    gtk_widget_destroy (dialog);

    if (response == GTK_RESPONSE_YES)
    {
        wk_view_stop_session (view);
        wk_daemon_handle_stop_session (d);
    }
}

static void
_view_clear_session (GtkWidget *widget, WkView *view)
{
    LOG ("%s", "CLEAR (USER)");
    wk_browser_clear_session (WK_BROWSER (wk_browser_get (NULL)), FALSE);
    _view_set_webpage (view, view->priv->settings->default_url);
}

static gboolean
_view_url_entry_modified (GtkWidget *entry, GdkEventKey *event, WkView *view)
{
    view->priv->entry_edited = true;

    return FALSE;
}

static void
_view_set_current_view (GtkWidget *entry, WkView *view)
{
    gchar *uri = NULL;

    view->priv->entry_edited = false;

    /* parse the uri for specific keywords */
    if ((g_strcmp0 (uri, "exit") == 0)
        || (g_strcmp0 (uri, "quit") == 0)
        || (g_strcmp0 (uri, "bye") == 0))
    {
        wk_browser_maybe_quit (WK_BROWSER (wk_browser_get (NULL)));
    }

    uri = expand_uri (gtk_entry_get_text (GTK_ENTRY (entry)));

    _view_set_history_buttons (view);
    _view_set_webpage (view, uri);

    g_free (uri);
}

static WebKitWebView *
_view_create_web_view (WebKitWebView  *web_view,
                       WebKitWebFrame *frame,
                       gpointer        data)
{
    GtkWidget *view;
    GtkWidget *window;
    WkSettings *settings;
    WkBrowser *browser;

    DEBUG ("start");

    settings = GET_SETTINGS();
    browser = WK_BROWSER (wk_browser_get (NULL));

    if ((settings->max_windows > 0)
        && (settings->max_windows <= wk_browser_get_windows_count (browser)))
    {
        DEBUG ("too much windows open");
        return NULL;
    }

    /* Create the window and notify the browser */
    window = wk_window_new (WK_WINDOW_TYPE_CHILD, NULL);
    wk_browser_register_window (browser, WK_WINDOW (window));
    view = wk_window_get_view (WK_WINDOW (window));

    DEBUG ("new view/window created");

    g_signal_connect (G_OBJECT (WK_VIEW(view)->webview), "notify::load-status",
                      G_CALLBACK (_view_on_popup_status_changed), window);

    DEBUG ("end");

    return WK_VIEW(view)->webview;
}

static void
_view_check_in_blacklist (gpointer data, gpointer user_data)
{
    const gchar *uri;
    const gchar *regex;

    uri = (const gchar *) user_data;
    regex = (const gchar *) data;

    if (g_regex_match_simple (regex, uri, G_REGEX_CASELESS, 0))
        is_granted = FALSE;
}

static void
_view_check_in_whitelist (gpointer data, gpointer user_data)
{
    const gchar *uri;
    const gchar *regex;

    uri = (const gchar *) user_data;
    regex = (const gchar *) data;

    if (g_regex_match_simple (regex, uri, G_REGEX_CASELESS, 0))
        is_granted = TRUE;
}

static gboolean
_view_uri_access_granted (const gchar *uri)
{
    WkSettings *settings;

    if (! uri)
        return TRUE;

    settings = GET_SETTINGS();
    is_granted = FALSE;

    DEBUG ("Testing access policy on %s", uri);

    if (g_str_has_prefix (uri, "file://")
        && ! settings->can_access_fs)
    {
        is_granted = FALSE;
        goto end;
    }

    if (settings->list_mode == LIST_MODE_BLACKLIST)
    {
        DEBUG ("using blacklist");
        is_granted = TRUE;
        g_list_foreach (settings->list, _view_check_in_blacklist, (gpointer) uri);
        goto end;
    }
    else if (settings->list_mode == LIST_MODE_WHITELIST)
    {
        DEBUG ("using whitelist");
        is_granted = FALSE;
        g_list_foreach (settings->list, _view_check_in_whitelist, (gpointer) uri);
        goto end;
    }

    is_granted = TRUE;

end:
    DEBUG ("Policy: %s", is_granted ? "Granted": "Denied");
    return is_granted;
}

static gboolean
_view_check_mime_policy (WebKitWebView           *web_view,
                         WebKitWebFrame          *frame,
                         WebKitNetworkRequest    *request,
                         gchar                   *mimetype,
                         WebKitWebPolicyDecision *policy_decision,
                         WkView                  *view)
{
    const gchar *uri;

    uri = webkit_network_request_get_uri (request);

    DEBUG ("mimetype check (%s) on uri %s", mimetype?mimetype:"", uri);

    if (strncmp (mimetype, "text/", 5) == 0)
    {
        DEBUG ("browser state: %d", view->priv->state);
        if (view->priv->state == WK_VIEW_STATE_DISPLAY)
        {
            webkit_web_policy_decision_use (policy_decision);
            return TRUE;
        }

        if (! _view_uri_access_granted (uri))
        {
            webkit_web_policy_decision_ignore (policy_decision);
            LOG ("URL (denied): %s", uri);
            DEBUG ("URL (denied): %s", uri);

            gtk_statusbar_pop (GTK_STATUSBAR (view->priv->statusbar), context_id);
            gtk_statusbar_push (GTK_STATUSBAR (view->priv->statusbar),
                                context_id, _("Access denied."));

            return TRUE;
        }

        LOG ("URL (allowed): %s", uri);
        DEBUG ("URL (allowed): %s", uri);

        webkit_web_policy_decision_use (policy_decision);
        return TRUE;
    }

    /* if mimetype is not text/ */
    if (view->priv->settings->mime_associations
        && g_hash_table_lookup (view->priv->settings->mime_associations,
                                mimetype) != NULL)
    {
        webkit_web_policy_decision_download (policy_decision);
        DEBUG ("downloading %s (%s)",
               webkit_network_request_get_uri (request),
               (gchar *) g_hash_table_lookup (
                            view->priv->settings->mime_associations,
                           mimetype));
        return TRUE;
    }

    return FALSE;
}

static void
_view_download_notify_status (WebKitDownload *download,
                              GParamSpec     *pspec,
                              WkView         *view)
{
    if (webkit_download_get_status (download) != WEBKIT_DOWNLOAD_STATUS_FINISHED)
        return;

    gchar *handler = NULL;
    gchar *filename = NULL;
    WebKitNetworkResponse *response;
    SoupMessage *message;
    const gchar *mimetype;
    gchar *cmd;
    int argc;
    gchar **argv;
    GPid pid;
    WkBrowser *browser;

    browser = WK_BROWSER (wk_browser_get (NULL));

    response = webkit_download_get_network_response (download);
    message = webkit_network_response_get_message (response);
    mimetype = soup_message_headers_get_content_type (message->response_headers, NULL);

    if (!view->priv->settings->mime_associations)
        return;

    handler = g_hash_table_lookup (view->priv->settings->mime_associations, mimetype);
    filename = g_filename_from_uri (webkit_download_get_destination_uri (download),
                                    NULL, NULL);

    cmd = g_strdup_printf ("%s '%s'", handler, filename);
    if (g_shell_parse_argv (cmd, &argc, &argv, NULL))
    {
        if (g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
                           NULL, NULL, &pid, NULL))
        {
            wk_browser_register_external (browser, pid);
        }
    }

    g_free (filename);
    g_strfreev (argv);
    g_free (cmd);
}

static gboolean
_view_do_download (WebKitWebView *web_view,
                   GObject       *download_obj,
                   WkView        *view)
{
    gchar *filename = NULL;
    gchar *uri = NULL;

    WebKitDownload *download = (WebKitDownload *) download_obj;

    filename = g_build_filename (g_get_tmp_dir (),
                                 view->priv->settings->tmpdir_name,
                                 webkit_download_get_suggested_filename (download),
                                 NULL);
    uri = g_filename_to_uri (filename, NULL, NULL);

    webkit_download_set_destination_uri (download, uri);
    g_signal_connect (download, "notify::status",
        G_CALLBACK (_view_download_notify_status), view);

    LOG ("DOWNLOAD: %s", uri);

    g_free (uri);
    g_free (filename);

    return TRUE;
}

static void
_view_build_statusbar (WkView *view)
{
    view->priv->statusbar = gtk_statusbar_new ();
    view->priv->progressbar = gtk_progress_bar_new ();
    gtk_box_pack_end (GTK_BOX (view->priv->statusbar), view->priv->progressbar,
                      FALSE, FALSE, 0);

    context_id = gtk_statusbar_get_context_id (
                    GTK_STATUSBAR (view->priv->statusbar), "Link Hover");
}

static void
_view_build_toolbar (WkView *view)
{
    GtkToolItem *item;

    view->priv->toolbar = gtk_toolbar_new ();
    gtk_toolbar_set_style (GTK_TOOLBAR (view->priv->toolbar),
                           GTK_TOOLBAR_ICONS);

    view->priv->backwardBtn = gtk_tool_button_new_from_stock (GTK_STOCK_GO_BACK);
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->backwardBtn), FALSE);
    gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->backwardBtn),
                                 _("Go back"));
    g_signal_connect (G_OBJECT (view->priv->backwardBtn), "clicked",
                      G_CALLBACK (_view_go_back), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                        view->priv->backwardBtn,
                        -1);

    view->priv->forwardBtn = gtk_tool_button_new_from_stock (GTK_STOCK_GO_FORWARD);
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->forwardBtn), FALSE);
    gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->forwardBtn),
                                 _("Go forward"));
    g_signal_connect (G_OBJECT (view->priv->forwardBtn), "clicked",
                      G_CALLBACK (_view_go_forward), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                        view->priv->forwardBtn,
                        -1);

    view->priv->homeBtn = gtk_tool_button_new_from_stock (GTK_STOCK_HOME);
    gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->forwardBtn),
                                 _("Go to home page"));
    g_signal_connect (G_OBJECT (view->priv->homeBtn), "clicked",
                      G_CALLBACK (_view_go_home), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                        view->priv->homeBtn,
                        -1);

    view->priv->stopBtn = gtk_tool_button_new_from_stock (GTK_STOCK_STOP);
    gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->forwardBtn),
                                 _("Stop loading"));
    gtk_widget_set_sensitive (GTK_WIDGET (view->priv->stopBtn), FALSE);
    g_signal_connect (G_OBJECT (view->priv->stopBtn), "clicked",
                      G_CALLBACK (_view_stop_loading), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                        view->priv->stopBtn,
                        -1);

    item = gtk_tool_item_new ();
    gtk_tool_item_set_expand (item, TRUE);
    view->priv->urlEntry = gtk_entry_new ();
    gtk_container_add (GTK_CONTAINER (item), view->priv->urlEntry);
    g_signal_connect (G_OBJECT (view->priv->urlEntry), "activate",
                      G_CALLBACK (_view_set_current_view), view);
    g_signal_connect (G_OBJECT (view->priv->urlEntry), "key-press-event",
                      G_CALLBACK (_view_url_entry_modified), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar), item, -1);

    item = gtk_tool_button_new_from_stock (GTK_STOCK_OK);
    gtk_widget_set_tooltip_text (GTK_WIDGET (item),
                                 _("Go to the address"));
    g_signal_connect (G_OBJECT (item), "clicked",
                      G_CALLBACK (_view_go), view);
    gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar), item, -1);

    if (view->priv->settings->with_timer)
    {
        item = gtk_tool_item_new ();
        view->priv->timeLabel = gtk_label_new ("00:00");
        gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->timeLabel),
                                     _("Remaining session time"));
        gtk_container_add (GTK_CONTAINER (item), view->priv->timeLabel);
        gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar), item, -1);
    }

    if (view->priv->settings->with_clear)
    {
        view->priv->clearBtn = gtk_tool_button_new_from_stock (GTK_STOCK_CLEAR);
        gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->clearBtn),
                                     _("Clear history and cookies"));
        g_signal_connect (G_OBJECT (view->priv->clearBtn), "clicked",
                          G_CALLBACK (_view_clear_session), view);
        gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                            view->priv->clearBtn,
                            -1);
    }

    if (view->priv->settings->can_end_session)
    {
        view->priv->resetBtn = gtk_tool_button_new_from_stock (GTK_STOCK_QUIT);
        gtk_widget_set_tooltip_text (GTK_WIDGET (view->priv->resetBtn),
                                     _("End this session"));
        g_signal_connect (G_OBJECT (view->priv->resetBtn), "clicked",
                          G_CALLBACK (_view_ask_stop_session), view);
        gtk_toolbar_insert (GTK_TOOLBAR (view->priv->toolbar),
                            view->priv->resetBtn,
                            -1);
    }
}

static gboolean
_view_set_image (WkView *view, const gchar *uri)
{
    const gchar *filename;
    GdkPixbuf *pxb;

    /* this stops possible media streams in webpages */
    _view_empty_webview (view);

    if (g_str_has_prefix (uri, "file://"))
        filename = uri+7;
    else
        filename = uri;

    if ((pxb = get_pixbuf_from_filename (filename, view)) == NULL)
        return FALSE;
    gtk_image_set_from_pixbuf (GTK_IMAGE (view->image), pxb);
    gtk_notebook_set_current_page (GTK_NOTEBOOK (view), VIEW_PAGE_IMAGE);
    gdk_flush ();

    g_object_unref (pxb);

    return TRUE;
}

static void
on_auth_dialog_response (GtkWidget *dialog, gint response, WkView *view)
{
    GtkEntry *user = g_object_get_data (G_OBJECT (dialog), "user");
    GtkEntry *password = g_object_get_data (G_OBJECT (dialog), "password");
    SoupMessage *msg = g_object_get_data (G_OBJECT (dialog), "msg");
    SoupAuth *auth = g_object_get_data (G_OBJECT (dialog), "auth");

    if (response == GTK_RESPONSE_ACCEPT)
    {
        soup_auth_authenticate (auth,
                                gtk_entry_get_text (user),
                                gtk_entry_get_text (password));
        gtk_widget_destroy (dialog);

        /* ugly temp workaround to correctly display the page... */
        webkit_web_view_load_uri (view->webview, view->priv->uri);
        return;
    }

    DEBUG ("Unpausing message");
    if (g_object_get_data (G_OBJECT (msg), "paused"))
    {
        soup_session_unpause_message (view->priv->session, msg);
        g_object_unref (msg);
    }
    gtk_widget_destroy (dialog);
}

static void
_view_on_auth_requested (SoupSession *session,
                         SoupMessage *msg,
                         SoupAuth *auth,
                         gboolean retrying,
                         WkView *view)
{
    SoupURI *uri;

    /* leave if the message is already marked (avoids double auth popup) */
    if (!retrying && g_object_get_data (G_OBJECT (msg), "ltop-mark"))
        return;

    g_object_set_data (G_OBJECT (msg), "ltop-mark", (void*)1);

    DEBUG ("Pausing message");
    soup_session_pause_message (session, msg);
    g_object_set_data (G_OBJECT (msg), "paused", (void*)1);
    g_object_ref (msg);

    uri = soup_uri_new (view->priv->uri);

    DEBUG ("auth on %s: %s / %s", view->priv->uri, uri->user, uri->password);
    DEBUG ("retrying: %s", retrying ? "yes" : "no");

    if (uri->user && uri->password)
    {
        soup_auth_authenticate (auth, uri->user, uri->password);
        DEBUG ("Unpausing message");
        if (g_object_get_data (G_OBJECT (msg), "paused"))
        {
            soup_session_unpause_message (session, msg);
            g_object_unref (msg);
        }
        return;
    }

    /* show a dialog for authentication */
    GtkWidget *d, *user_entry, *password_entry;
    GtkWidget *hbox, *label;
    GtkSizeGroup *sg;

    d = gtk_dialog_new_with_buttons (_("Authentication required"),
                                     NULL, GTK_DIALOG_MODAL,
                                     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                     GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
                                     NULL);
    gtk_dialog_set_default_response (GTK_DIALOG (d), GTK_RESPONSE_ACCEPT);
    gtk_container_set_border_width (GTK_CONTAINER (d), 6);
    gtk_window_set_icon_name (GTK_WINDOW (d), GTK_STOCK_DIALOG_AUTHENTICATION);

    label = gtk_label_new (
                _("The website you're browsing requires authentication:\n"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), label, FALSE, FALSE, 2);

    sg = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

    hbox = gtk_hbox_new (FALSE, 0);
    label = gtk_label_new (_("Username: "));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_size_group_add_widget (sg, label);
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
    user_entry = gtk_entry_new ();
    gtk_box_pack_start (GTK_BOX (hbox), user_entry, TRUE, TRUE, 2);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), hbox, FALSE, FALSE, 2);

    hbox = gtk_hbox_new (FALSE, 0);
    label = gtk_label_new (_("Password: "));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_size_group_add_widget (sg, label);
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
    password_entry = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (password_entry), FALSE);
    gtk_entry_set_activates_default (GTK_ENTRY (password_entry), TRUE);
    gtk_box_pack_start (GTK_BOX (hbox), password_entry, TRUE, TRUE, 2);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), hbox, FALSE, FALSE, 2);

    g_object_set_data (G_OBJECT (d), "user", user_entry);
    g_object_set_data (G_OBJECT (d), "password", password_entry);
    g_object_set_data (G_OBJECT (d), "msg", msg);
    g_object_set_data (G_OBJECT (d), "auth", auth);
    g_signal_connect (G_OBJECT (d), "response",
                      G_CALLBACK (on_auth_dialog_response), view);
    gtk_widget_show_all (d);
}

static void
_view_on_request_queued (SoupSession *session,
                         SoupMessage *msg,
                         WkView      *view)
{
    g_signal_connect (G_OBJECT (session), "authenticate",
                      G_CALLBACK (_view_on_auth_requested), view);
    g_signal_handlers_disconnect_by_func (G_OBJECT (session),
                                          _view_on_request_queued, view);

}

static gboolean
_view_set_webpage (WkView *view, const gchar *uri)
{
    g_signal_connect (G_OBJECT (view->priv->session), "request-queued",
                      G_CALLBACK (_view_on_request_queued), view);

    g_free (view->priv->uri);
    view->priv->uri = g_strdup (uri);

    webkit_web_view_load_uri (view->webview, uri);

    return TRUE;
}

static void
_view_reinit_video_tab (WkView *view)
{
    gtk_widget_destroy (view->video);

    view->video = wk_video_new (view->priv->settings->slideshow_mute);
    g_signal_connect (G_OBJECT (view->video), "end-reached",
                      G_CALLBACK (_view_on_video_end), view);
    gtk_notebook_remove_page (GTK_NOTEBOOK (view), VIEW_PAGE_VIDEO);
    gtk_notebook_insert_page (GTK_NOTEBOOK (view), view->video,
                              NULL, VIEW_PAGE_VIDEO);
    gtk_widget_show (view->video);
}

static gboolean
_view_after_video (WkView *view)
{
    _view_reinit_video_tab (view);
    wk_daemon_handle_video_end (wk_daemon_get ());

    return FALSE;
}

static void
_view_on_video_end (WkVideo *video, WkView *view)
{
    g_timeout_add (100, (GSourceFunc) _view_after_video, view);
}

static gboolean
_view_set_video (WkView *view, const gchar *uri)
{
    gboolean ret;

    /* this stops possible media streams in webpages */
    _view_empty_webview (view);

    ret = wk_video_play (WK_VIDEO (view->video), uri);
    gtk_notebook_set_current_page (GTK_NOTEBOOK (view), VIEW_PAGE_VIDEO);
    gdk_flush ();

    return ret;
}

static gboolean
_view_decrease_time (WkView *view)
{
    if (G_LIKELY (--(view->priv->time) > 0)) {
        gchar *time_str = NULL;
        seconds_to_string (view->priv->time, &time_str);
        gtk_label_set_text (GTK_LABEL (view->priv->timeLabel), time_str);
        g_free (time_str);
        return TRUE;
    }

    g_signal_emit (G_OBJECT (view), fe_signals[SESSION_STOP_SIGNAL], 0);
    WkDaemon *d = wk_daemon_get ();
    wk_daemon_handle_stop_session (d);

    return FALSE;
}

/*
 * Public API
 */
GtkWidget *
wk_view_new ()
{
    WkView *view = g_object_new (WK_TYPE_VIEW, NULL);
    return GTK_WIDGET (view);
}

WkViewState
wk_view_get_state (WkView *view)
{
    return view->priv->state;
}

gboolean
wk_view_show (WkView *view, guint type, const gchar *uri)
{
    SoupCookieJar *jar;
    gboolean ret;

    DEBUG ("entering");
    view->priv->state = WK_VIEW_STATE_DISPLAY;

    webkit_web_view_stop_loading (view->webview);
    wk_view_stop_session (view);
    wk_video_stop (WK_VIDEO (view->video));

    /* set the cookies */
    if (0 == g_access (view->priv->settings->cookies_file, R_OK))
    {
        jar = soup_cookie_jar_sqlite_new (view->priv->settings->cookies_file,
                                          TRUE);
        soup_session_add_feature (view->priv->session,
                                  SOUP_SESSION_FEATURE (jar));
    }

    if (type == WK_SLIDESHOW_ITEM_TYPE_VIDEO)
    {
        DEBUG ("setting up video %s", uri);
        ret = _view_set_video (view, uri);
    }
    else if (type == WK_SLIDESHOW_ITEM_TYPE_IMAGE)
    {
        DEBUG ("setting up image %s", uri);
        ret = _view_set_image (view, uri);
    }
    else
    {
        DEBUG ("setting up webpage %s", uri);
        ret = _view_set_webpage (view, uri);
    }

    return ret;
}

void
wk_view_set_url (WkView *view, gchar *url)
{
    _view_set_webpage (view, url);
}

void
wk_view_show_statusbar (WkView *view, gboolean show)
{
    if (show)
        gtk_widget_show (view->priv->statusbar);
    else
        gtk_widget_hide (view->priv->statusbar);
}

void
wk_view_show_toolbar (WkView *view, gboolean show)
{
    if (show && ! view->priv->settings->hide_toolbar)
        gtk_widget_show (view->priv->toolbar);
    else
        gtk_widget_hide (view->priv->toolbar);
}

guint
wk_view_get_time (WkView *view)
{
    return view->priv->time;
}

void
wk_view_start_session (WkView *view)
{
    WebKitWebBackForwardList *list;

    LOG ("%s", "START");
    DEBUG ("starting session");

    stop_unclutter ();
    wk_video_stop (WK_VIDEO (view->video));

    gdk_event_handler_set (event_no_filter, NULL, NULL);

    if (view->priv->state == WK_VIEW_STATE_WEB)
        return;

    view->priv->state = WK_VIEW_STATE_WEB;

    DEBUG ("connecting to navigation policy signal");
    g_signal_connect (view->webview, "mime-type-policy-decision-requested",
                      G_CALLBACK (_view_check_mime_policy), view);

    wk_view_show_toolbar (view, TRUE);
    wk_view_show_statusbar (view, TRUE);

    list = webkit_web_back_forward_list_new_with_web_view (view->webview);
    webkit_web_back_forward_list_set_limit (list,
                                            view->priv->settings->max_history);

    soup_session_remove_feature_by_type (view->priv->session,
                                      g_type_from_name ("WebKitSoupAuthDialog"));

    soup_session_add_feature_by_type (view->priv->session,
                                      soup_cookie_jar_get_type ());

    _view_set_webpage (view, view->priv->settings->default_url);
    _view_set_history_buttons (view);

    if (view->priv->settings->with_timer)
    {
        wk_view_add_time (view, view->priv->settings->default_timer);
        view->priv->time_timeout = g_timeout_add_seconds (
                                    1, (GSourceFunc) _view_decrease_time,
                                    view);
    }

    g_signal_emit (G_OBJECT (view), fe_signals[SESSION_START_SIGNAL], 0);

    DEBUG ("executing hook session-start");
    exec_hook (view->priv->settings->hook_dir, "session-start");
}

void
wk_view_stop_session (WkView *view)
{
    WebKitWebBackForwardList *list;

    if (view->priv->state != WK_VIEW_STATE_WEB)
        return;

    LOG ("%s", "STOP");

    g_signal_handlers_disconnect_by_func (view->webview, _view_check_mime_policy,
                                          view);

    view->priv->time = 0;
    if (view->priv->time_timeout)
    {
        g_source_destroy (g_main_context_find_source_by_id (
            NULL, view->priv->time_timeout));
        view->priv->time_timeout = 0;
        gtk_label_set_text (GTK_LABEL (view->priv->timeLabel), "00:00");
    }

    soup_session_remove_feature_by_type (view->priv->session,
                                         soup_cookie_jar_get_type ());
    soup_session_add_feature_by_type (view->priv->session,
                                      soup_cookie_jar_get_type ());

    /* set the history limit to 0 to forget the previous user session */
    list = webkit_web_back_forward_list_new_with_web_view (view->webview);
    webkit_web_back_forward_list_set_limit (list, 0);

    view->priv->state = WK_VIEW_STATE_NONE;

    wk_view_show_toolbar (view, FALSE);
    wk_view_show_statusbar (view, FALSE);

    exec_hook (view->priv->settings->hook_dir, "session-stop");

    g_signal_emit (G_OBJECT (view), fe_signals[SESSION_STOP_SIGNAL], 0);
}

void
wk_view_add_time (WkView *view, guint seconds)
{
    g_return_if_fail (WK_IS_VIEW (view));

    if (!view->priv->settings->with_timer)
        return;

    gchar *time_str = NULL;

    view->priv->time += seconds;
    g_signal_emit (G_OBJECT (view), fe_signals[TIME_ADDED_SIGNAL], 0);

    seconds_to_string (view->priv->time, &time_str);
    gtk_label_set_text (GTK_LABEL (view->priv->timeLabel), time_str);
}

void
wk_view_clear_history (WkView *view)
{
    WebKitWebBackForwardList *list;

    soup_session_remove_feature_by_type (view->priv->session,
                                         soup_cookie_jar_get_type ());
    soup_session_add_feature_by_type (view->priv->session,
                                      soup_cookie_jar_get_type ());

    /* set the history limit to 0 to forget the previous user session */
    list = webkit_web_back_forward_list_new_with_web_view (view->webview);
    webkit_web_back_forward_list_set_limit (list, 0);
    webkit_web_back_forward_list_set_limit (list,
                                            view->priv->settings->max_history);

    _view_set_history_buttons (view);
}

void
wk_view_url_entry_grab_focus (WkView *view)
{
    gtk_widget_grab_focus (view->priv->urlEntry);
}

