/**
 * Copyright (C) 2007-2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <glib/gstdio.h>

#include <gtk/gtk.h>

#include "wk-key-file.h"

#include "xfce-heading.h"

#include "config-main.h"
#include "config-kiosk.h"
#include "config-display.h"
#ifdef HAVE_REMOTE
#include "config-remote.h"
#endif

#include "main.h"
#include "utils.h"

GtkWidget *kiosk_widget;
GtkWidget *display_widget;
GtkWidget *main_window;

GHashTable *main_hash, *kiosk_hash, *display_hash;

#ifdef HAVE_REMOTE
GtkWidget  *remote_widget;
GHashTable *remote_hash;
#endif

WkKeyFile *kf;

void
saveconf ()
{
    gchar *preffile = NULL;
    gchar *preftempfile = NULL;

    preffile = wk_key_file_build_filename (kf, "preferences");
    preftempfile = g_strdup_printf ("%s.tmp", preffile);

    config_kiosk_save_config (kf, kiosk_hash);
    config_display_save_config (kf, display_hash);
#ifdef HAVE_REMOTE
    config_remote_save_config (kf, remote_hash);
#endif

    /* MUST be handled at the end to override the start_with_session setting if
       needed */
    config_main_save_config (kf, main_hash);

    if (!wk_key_file_write (kf, preftempfile)
        || (g_rename (preftempfile, preffile) != 0))
    {
        /* TODO: we should have a way to save the config in an other file */
        gtk_dialog_run (GTK_DIALOG (gtk_message_dialog_new (
            NULL, 0, GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            _("Unable to save the configuration file "
            "to\n%s"), preffile)));
    }

    g_free (preffile);
    g_free (preftempfile);
}

static void
on_help_btn_clicked (GtkWidget *widget, gpointer data)
{
    g_spawn_command_line_async ("yelp ghelp:linutop-webkiosk",
                                NULL);
}

static void
on_test_btn_clicked (GtkWidget *widget, gpointer data)
{
    gchar *tmp;
    gchar *cmd;
    gchar *out, *err;
    gchar *oldbase;

    tmp = g_strdup_printf ("/tmp/webkiosk-test-%d", getpid ());

    oldbase = g_strdup (wk_key_file_get_basedir (kf));

    cmd = g_strdup_printf ("cp %s/coockies.sqlite %s", oldbase, tmp);
    g_spawn_command_line_sync (cmd, &out, &err, NULL, NULL);
    g_free (cmd);
    g_free (out);
    g_free (err);

    cmd = g_strdup_printf ("cp -R %s/hooks %s", oldbase, tmp);
    g_spawn_command_line_sync (cmd, &out, &err, NULL, NULL);
    g_free (cmd);
    g_free (out);
    g_free (err);

    wk_key_file_set_basedir (kf, tmp);
    saveconf ();

    cmd = g_strdup_printf ("linutop-webkiosk -c %s", tmp);
    g_spawn_command_line_sync (cmd, &out, &err, NULL, NULL);
    g_free (cmd);
    g_free (out);
    g_free (err);

    wk_key_file_set_basedir (kf, (gchar *) oldbase);

    empty_dir (tmp);
    g_rmdir (tmp);
    g_free (tmp);
    g_free (oldbase);
}

static void
on_dialog_response (GtkDialog *dialog, gint response_id, gpointer data)
{
    switch (response_id)
    {
        case GTK_RESPONSE_OK:
            saveconf ();
        case GTK_RESPONSE_REJECT:
        case GTK_RESPONSE_DELETE_EVENT:
            gtk_main_quit ();
            break;
    }
}

int
main (int argc, gchar **argv)
{
    GtkWidget *notebook;
    GtkWidget *heading;
    GtkWidget *test_btn, *help_btn;
    GdkPixbuf *pxb;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
    textdomain(GETTEXT_PACKAGE);

    if (! g_thread_supported ())
        g_thread_init (NULL);
    gtk_init (&argc, &argv);

    /* init the config file */
    kf = wk_key_file_new (NULL);
    wk_key_file_load (kf);

    /* init the interface */
    gtk_window_set_default_icon_from_file (WEBKIOSKICON, NULL);

    main_window = gtk_dialog_new_with_buttons (
        _("Linutop Kiosk - Setup"), NULL, 0, NULL);

    gtk_dialog_set_has_separator (GTK_DIALOG (main_window), FALSE);

    test_btn = gtk_dialog_add_button (GTK_DIALOG (main_window),
                                      _("Run a test"), -1);
    gtk_button_set_image (GTK_BUTTON (test_btn),
        gtk_image_new_from_stock (GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON));
    g_signal_connect (G_OBJECT (test_btn), "clicked",
                      G_CALLBACK (on_test_btn_clicked), NULL);

    gtk_dialog_add_button (GTK_DIALOG (main_window), GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT);
    gtk_dialog_add_button (GTK_DIALOG (main_window), GTK_STOCK_OK, GTK_RESPONSE_OK);

    help_btn = gtk_dialog_add_button (GTK_DIALOG (main_window),
                                      GTK_STOCK_HELP, -1);
    gtk_button_box_set_child_secondary (
        GTK_BUTTON_BOX (GTK_DIALOG (main_window)->action_area),
        help_btn, TRUE);
    g_signal_connect (G_OBJECT (help_btn), "clicked",
                      G_CALLBACK (on_help_btn_clicked), NULL);

    heading = xfce_heading_new ();
    pxb = gdk_pixbuf_new_from_file (WEBKIOSKICON, NULL);
    if (pxb)
    {
        xfce_heading_set_icon (XFCE_HEADING (heading), pxb);
        g_object_unref (pxb);
    }
    xfce_heading_set_title (XFCE_HEADING (heading),
                            _("Linutop Kiosk - Setup"));
    xfce_heading_set_subtitle (XFCE_HEADING (heading),
                               _("For Secure Internet Access and Digital Signage"));

    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (main_window)->vbox),
        heading, FALSE, FALSE, 2);

    notebook = gtk_notebook_new ();
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (main_window)->vbox),
        notebook, TRUE, TRUE, 2);

    /* TODO: automate registration of pages? */
    main_hash = config_main_get_hash (kf);
    kiosk_hash = config_kiosk_get_hash (kf);
    display_hash = config_display_get_hash (kf);
#ifdef HAVE_REMOTE
    remote_hash = config_remote_get_hash (kf);
#endif

    kiosk_widget = config_kiosk_get_widget (kiosk_hash);
    gtk_widget_set_visible (kiosk_widget, FALSE);
    display_widget = config_display_get_widget (display_hash);
#ifdef HAVE_REMOTE
    remote_widget = config_remote_get_widget (remote_hash);
#endif

    /* the main page is handled differently from the others because it acts on
       the notebook directly */
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook),
                              config_main_get_widget (notebook),
                              gtk_label_new (_("General")));

    gtk_notebook_append_page (GTK_NOTEBOOK (notebook),
                              kiosk_widget,
                              gtk_label_new (_("Web Kiosk")));
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook),
                              display_widget,
                              gtk_label_new (_("Digital Signage / Screen Saver")));
#ifdef HAVE_REMOTE
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook),
                              remote_widget,
                              gtk_label_new (_("Remote Control")));
#endif

    gtk_widget_show (heading);
    gtk_widget_show (notebook);
    gtk_widget_show (main_window);

    g_object_set (gtk_settings_get_default (),
                  "gtk-button-images", TRUE, NULL);

    g_signal_connect (G_OBJECT (main_window), "response",
                      G_CALLBACK (on_dialog_response), NULL);
    gtk_main ();

    /* cleanup */
    g_object_unref (kf);
    g_hash_table_destroy (kiosk_hash);
    g_hash_table_destroy (display_hash);
#ifdef HAVE_REMOTE
    g_hash_table_destroy (remote_hash);
#endif
    g_hash_table_destroy (main_hash);

    return 0;
}

